import csv, os, tabulate
from enum import Enum

soclabs_header = """;#-----------------------------------------------------------------------------
;# SoC Labs Basic Hashing Accelerator Wrapper Input Stimulus File
;# A joint work commissioned on behalf of SoC Labs, under Arm Academic Access license.
;#
;# Contributors
;#
;# David Mapstone (d.a.mapstone@soton.ac.uk)
;#
;# Copyright  2023, SoC Labs (www.soclabs.org)
;#-----------------------------------------------------------------------------"""

class TransactionType(Enum):
    """ Enumerated Types for Transaction Types for ASCII Debug """
    READ  = 1
    WRITE = 2
    def __str__(self):
        match self:
            case TransactionType.READ:
                return "R"
            case TransactionType.WRITE:
                return "W"

class TransactionSize(Enum):
    """ Enumerated Types for Transaction Types for ASCII Debug """
    WORD     = 1
    HALFWORD = 2
    def __str__(self):
        match self:
            case TransactionSize.WORD:
                return "word"
            case TransactionSize.HALFWORD:
                return "halfword"

class InputBlockStruct:
    def __init__(self):
        self.word_list = []
    
    def word_append(self, word):
        self.word_list.append(word)

class InputPacketStruct:
    def __init__(self):
        self.block_list = []
    
    def block_append(self, block):
        self.block_list.append(block)

class WordStruct:
    def __init__(self, data, addr, trans, packet_num = 0, block_num = 0, size = TransactionSize.WORD):
        self.data = data
        self.addr = addr
        self.trans = trans
        self.packet_num = packet_num
        self.block_num = block_num
        self.size = size

def fri_output(out_file, word_list):
    """ 
    This function takes a list of 32 bit words and addresses and formats 
    the data into .fri format to be fed into fml2conv.pl script to stimulate
    testbench
    """
    
    # Column Names
    col_names = ["Transaction", "Address", "Data", "Size"]

    data = []
    for word in word_list:
        if (word.data > 0):
            data.append([str(word.trans), "{0:#0{1}x}".format(word.addr,10), "{0:#0{1}x}".format(word.data,10), str(word.size)])
    
    table_str = tabulate.tabulate(data, headers=col_names, tablefmt="plain")

    with open(out_file, "w", encoding="UTF8", newline='') as f:
        f.write(soclabs_header + "\n;")
        f.write(table_str)
        f.write("\nQ") # Add End of Simulation Flag

def stimulus_generation(in_file, start_address, size):
    """ 
    This function takes 32 bit input stimulus file from accelerator model,
    calculates write addresses for each word and generates a .fri file which
    can be used to stimulate an AHB testbench
    """
    # Calculate End Address
    end_address = start_address + size - 0x4
    # print(f"End Address is {hex(end_address)}")

    # Open File
    with open(in_file, "r") as stim:
        csvreader = csv.reader(stim, delimiter=",")
        stim_list = list(csvreader)

    # Initialise Packet List
    packet_list = []

    # Initialise Temporary Structures
    temppacketstruct = InputPacketStruct()
    tempblockstruct = InputBlockStruct()

    # Read Data into Structs
    for i in stim_list:
        tempblockstruct.word_append(int(i[0],16))
        # If Last Word in Block, Append to Packet and Reset Temporary block structure
        if (int(i[1])):
            temppacketstruct.block_append(tempblockstruct)
            tempblockstruct = InputBlockStruct()
            # If Last Block in Packet , Append Packet to Packet List and Reset Temp Packet
            if (int(i[2])):
                packet_list.append(temppacketstruct)
                temppacketstruct = InputPacketStruct()
        
    
    # List of Ouptut Transactions
    output_word_list = []

    # Generate Address for Packet
    for packet_num, packet in enumerate(packet_list):
        # Calculate Number of Blocks in First Packet
        num_blocks = len(packet.block_list)
        # Each Block Can Contain 16 32-bit Words (512 bits) (0x4 * 16 = 0x40)
        # - Work Out Required Size = (0x40 * NumBlocks)
        # - Work Out Beginning Address = (end_address + 0x4) - Size
        req_write_size = 0x40 * num_blocks
        start_write_addr = start_address + size - req_write_size
        # print(f"Packet: {int(packet_num)} | Start Address: {hex(start_write_addr)}")
        write_addr = start_write_addr
        for block_num, block in enumerate(packet.block_list):
            read_address = write_addr + 0x0000_0800
            for word in block.word_list:
                word_data = WordStruct(word, write_addr, TransactionType.WRITE, packet_num, block_num)
                output_word_list.append(word_data)
                # Increment Address
                write_addr += 0x4
            for word in block.word_list:
                word_data = WordStruct(word, read_address, TransactionType.READ, packet_num, block_num)
                output_word_list.append(word_data)
                # Increment Address
                read_address += 0x4

    # Generate FRI File with Write Transactions
    fri_file = os.environ["WRAP_ACC_DIR"] + "/simulate/stimulus/" + "ahb_input_hash_stim.fri"
    fri_output(fri_file, output_word_list)

    # Call fm2conv.pl script
    m2d_file = os.environ["WRAP_ACC_DIR"] + "/simulate/stimulus/" + "ahb_input_hash_stim.m2d"
    os.system(f"fm2conv.pl -busWidth=32 -infile={fri_file} -outfile={m2d_file}")


if __name__ == "__main__":
    accelerator_input_address = 0x6001_0000
    accelerator_input_size = 0x0000_0800
    in_file = os.environ["SHA_2_ACC_DIR"] + "/simulate/stimulus/system/" + "input_data_32bit_stim.csv"
    stimulus_generation(in_file, accelerator_input_address, accelerator_input_size)